// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::{Array, ArrayType},
    bundle::{Bundle, BundleType},
    clock::Clock,
    enum_::{Enum, EnumType},
    expr::{CastBitsTo, Expr, ToExpr},
    int::{Bool, IntType, KnownSize, SInt, SIntType, SIntValue, Size, UInt, UIntType, UIntValue},
    reset::{AsyncReset, Reset, SyncReset},
    ty::{CanonicalType, Type},
    util::{
        alternating_cell::{AlternatingCell, AlternatingCellMethods},
        ConstUsize,
    },
};
use bitvec::{slice::BitSlice, vec::BitVec};
use std::{
    fmt,
    ops::{Deref, DerefMut},
    sync::Arc,
};

#[derive(Copy, Clone, Eq, PartialEq)]
enum ValidFlags {
    BothValid = 0,
    OnlyValueValid = 1,
    OnlyBitsValid = 2,
}

#[derive(Clone)]
struct SimValueInner<T: Type> {
    value: T::SimValue,
    bits: UIntValue,
    valid_flags: ValidFlags,
    ty: T,
}

impl<T: Type> SimValueInner<T> {
    fn fill_bits(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid | ValidFlags::OnlyBitsValid => {}
            ValidFlags::OnlyValueValid => {
                self.ty.sim_value_to_bits(&self.value, self.bits.bits_mut());
                self.valid_flags = ValidFlags::BothValid;
            }
        }
    }
    fn into_bits(mut self) -> UIntValue {
        self.fill_bits();
        self.bits
    }
    fn bits_mut(&mut self) -> &mut UIntValue {
        self.fill_bits();
        self.valid_flags = ValidFlags::OnlyBitsValid;
        &mut self.bits
    }
    fn fill_value(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid | ValidFlags::OnlyValueValid => {}
            ValidFlags::OnlyBitsValid => {
                self.ty
                    .sim_value_clone_from_bits(&mut self.value, self.bits.bits());
                self.valid_flags = ValidFlags::BothValid;
            }
        }
    }
    fn into_value(mut self) -> T::SimValue {
        self.fill_value();
        self.value
    }
    fn value_mut(&mut self) -> &mut T::SimValue {
        self.fill_value();
        self.valid_flags = ValidFlags::OnlyValueValid;
        &mut self.value
    }
}

impl<T: Type> AlternatingCellMethods for SimValueInner<T> {
    fn unique_to_shared(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid => return,
            ValidFlags::OnlyValueValid => {
                self.ty.sim_value_to_bits(&self.value, self.bits.bits_mut())
            }
            ValidFlags::OnlyBitsValid => self
                .ty
                .sim_value_clone_from_bits(&mut self.value, self.bits.bits()),
        }
        self.valid_flags = ValidFlags::BothValid;
    }

    fn shared_to_unique(&mut self) {}
}

pub struct SimValue<T: Type> {
    inner: AlternatingCell<SimValueInner<T>>,
}

impl<T: Type + Clone> Clone for SimValue<T> {
    fn clone(&self) -> Self {
        Self {
            inner: AlternatingCell::new_unique(self.inner.share().clone()),
        }
    }
}

impl<T: Type> SimValue<T> {
    #[track_caller]
    pub fn from_bits(ty: T, bits: UIntValue) -> Self {
        assert_eq!(ty.canonical().bit_width(), bits.width());
        let inner = SimValueInner {
            value: ty.sim_value_from_bits(bits.bits()),
            bits,
            valid_flags: ValidFlags::BothValid,
            ty,
        };
        Self {
            inner: AlternatingCell::new_shared(inner),
        }
    }
    #[track_caller]
    pub fn from_bitslice(ty: T, bits: &BitSlice) -> Self {
        Self::from_bits(ty, UIntValue::new(Arc::new(bits.to_bitvec())))
    }
    pub fn from_value(ty: T, value: T::SimValue) -> Self {
        let inner = SimValueInner {
            bits: UIntValue::new_dyn(Arc::new(BitVec::repeat(false, ty.canonical().bit_width()))),
            value,
            valid_flags: ValidFlags::OnlyValueValid,
            ty,
        };
        Self {
            inner: AlternatingCell::new_unique(inner),
        }
    }
    pub fn ty(this: &Self) -> T {
        this.inner.share().ty
    }
    pub fn into_bits(this: Self) -> UIntValue {
        this.inner.into_inner().into_bits()
    }
    pub fn into_ty_and_bits(this: Self) -> (T, UIntValue) {
        let inner = this.inner.into_inner();
        (inner.ty, inner.into_bits())
    }
    pub fn bits(this: &Self) -> &UIntValue {
        &this.inner.share().bits
    }
    pub fn bits_mut(this: &mut Self) -> &mut UIntValue {
        this.inner.unique().bits_mut()
    }
    pub fn into_value(this: Self) -> T::SimValue {
        this.inner.into_inner().into_value()
    }
    pub fn value(this: &Self) -> &T::SimValue {
        &this.inner.share().value
    }
    pub fn value_mut(this: &mut Self) -> &mut T::SimValue {
        this.inner.unique().value_mut()
    }
    #[track_caller]
    pub fn from_canonical(v: SimValue<CanonicalType>) -> Self {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        Self::from_bits(T::from_canonical(ty), bits)
    }
    pub fn into_canonical(this: Self) -> SimValue<CanonicalType> {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(ty.canonical(), bits)
    }
    pub fn canonical(this: &Self) -> SimValue<CanonicalType> {
        SimValue::from_bits(Self::ty(this).canonical(), Self::bits(this).clone())
    }
    #[track_caller]
    pub fn from_dyn_int(v: SimValue<T::Dyn>) -> Self
    where
        T: IntType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_dyn_int(ty), bits)
    }
    pub fn into_dyn_int(this: Self) -> SimValue<T::Dyn>
    where
        T: IntType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(ty.as_dyn_int(), bits)
    }
    pub fn to_dyn_int(this: &Self) -> SimValue<T::Dyn>
    where
        T: IntType,
    {
        SimValue::from_bits(Self::ty(this).as_dyn_int(), Self::bits(&this).clone())
    }
    #[track_caller]
    pub fn from_bundle(v: SimValue<Bundle>) -> Self
    where
        T: BundleType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_canonical(CanonicalType::Bundle(ty)), bits)
    }
    pub fn into_bundle(this: Self) -> SimValue<Bundle>
    where
        T: BundleType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(Bundle::from_canonical(ty.canonical()), bits)
    }
    pub fn to_bundle(this: &Self) -> SimValue<Bundle>
    where
        T: BundleType,
    {
        SimValue::from_bits(
            Bundle::from_canonical(Self::ty(this).canonical()),
            Self::bits(&this).clone(),
        )
    }
    #[track_caller]
    pub fn from_enum(v: SimValue<Enum>) -> Self
    where
        T: EnumType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_canonical(CanonicalType::Enum(ty)), bits)
    }
    pub fn into_enum(this: Self) -> SimValue<Enum>
    where
        T: EnumType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(Enum::from_canonical(ty.canonical()), bits)
    }
    pub fn to_enum(this: &Self) -> SimValue<Enum>
    where
        T: EnumType,
    {
        SimValue::from_bits(
            Enum::from_canonical(Self::ty(this).canonical()),
            Self::bits(&this).clone(),
        )
    }
}

impl<T: Type> Deref for SimValue<T> {
    type Target = T::SimValue;

    fn deref(&self) -> &Self::Target {
        Self::value(self)
    }
}

impl<T: Type> DerefMut for SimValue<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        Self::value_mut(self)
    }
}

impl<T: Type> fmt::Debug for SimValue<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let inner = self.inner.share();
        f.debug_struct("SimValue")
            .field("ty", &inner.ty)
            .field("value", &inner.value)
            .finish()
    }
}

impl<T: Type> ToExpr for SimValue<T> {
    type Type = T;

    #[track_caller]
    fn to_expr(&self) -> Expr<Self::Type> {
        let inner = self.inner.share();
        inner.bits.cast_bits_to(inner.ty)
    }
}

pub trait SimValuePartialEq<T: Type = Self>: Type {
    #[track_caller]
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<T>) -> bool;
}

impl<T: SimValuePartialEq<U>, U: Type> PartialEq<SimValue<U>> for SimValue<T> {
    #[track_caller]
    fn eq(&self, other: &SimValue<U>) -> bool {
        T::sim_value_eq(self, other)
    }
}

impl<LhsWidth: Size, RhsWidth: Size> SimValuePartialEq<UIntType<RhsWidth>> for UIntType<LhsWidth> {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<UIntType<RhsWidth>>) -> bool {
        **this == **other
    }
}

impl<LhsWidth: Size, RhsWidth: Size> SimValuePartialEq<SIntType<RhsWidth>> for SIntType<LhsWidth> {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<SIntType<RhsWidth>>) -> bool {
        **this == **other
    }
}

impl SimValuePartialEq<Bool> for Bool {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<Bool>) -> bool {
        **this == **other
    }
}

pub trait ToSimValue<T: Type> {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T>;
    #[track_caller]
    fn into_sim_value(self, ty: T) -> SimValue<T>
    where
        Self: Sized,
    {
        self.to_sim_value(ty)
    }
    #[track_caller]
    fn arc_into_sim_value(self: Arc<Self>, ty: T) -> SimValue<T> {
        self.to_sim_value(ty)
    }
    #[track_caller]
    fn arc_to_sim_value(self: &Arc<Self>, ty: T) -> SimValue<T> {
        self.to_sim_value(ty)
    }
}

impl<T: Type> ToSimValue<T> for SimValue<T> {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        assert_eq!(SimValue::ty(self), ty);
        self.clone()
    }

    #[track_caller]
    fn into_sim_value(self, ty: T) -> SimValue<T> {
        assert_eq!(SimValue::ty(&self), ty);
        self
    }
}

impl<T: Type> ToSimValue<T> for BitVec {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        self.clone().into_sim_value(ty)
    }

    #[track_caller]
    fn into_sim_value(self, ty: T) -> SimValue<T> {
        Arc::new(self).arc_into_sim_value(ty)
    }

    #[track_caller]
    fn arc_into_sim_value(self: Arc<Self>, ty: T) -> SimValue<T> {
        SimValue::from_bits(ty, UIntValue::new_dyn(self))
    }

    #[track_caller]
    fn arc_to_sim_value(self: &Arc<Self>, ty: T) -> SimValue<T> {
        SimValue::from_bits(ty, UIntValue::new_dyn(self.clone()))
    }
}

impl<T: Type> ToSimValue<T> for bitvec::boxed::BitBox {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        self.clone().into_sim_value(ty)
    }

    #[track_caller]
    fn into_sim_value(self, ty: T) -> SimValue<T> {
        self.into_bitvec().into_sim_value(ty)
    }
}

impl<T: Type> ToSimValue<T> for BitSlice {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        self.to_bitvec().into_sim_value(ty)
    }
}

impl<This: ?Sized + ToSimValue<T>, T: Type> ToSimValue<T> for &'_ This {
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        This::to_sim_value(self, ty)
    }
}

impl<This: ?Sized + ToSimValue<T>, T: Type> ToSimValue<T> for &'_ mut This {
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        This::to_sim_value(self, ty)
    }
}

impl<This: ?Sized + ToSimValue<T>, T: Type> ToSimValue<T> for Arc<This> {
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        This::arc_to_sim_value(self, ty)
    }
    fn into_sim_value(self, ty: T) -> SimValue<T> {
        This::arc_into_sim_value(self, ty)
    }
}

impl<This: ?Sized + ToSimValue<T> + Send + Sync + 'static, T: Type> ToSimValue<T>
    for crate::intern::Interned<This>
{
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        This::to_sim_value(self, ty)
    }
}

impl<This: ToSimValue<T>, T: Type> ToSimValue<T> for Box<This> {
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        This::to_sim_value(self, ty)
    }
    fn into_sim_value(self, ty: T) -> SimValue<T> {
        This::into_sim_value(*self, ty)
    }
}

impl<T: Type, Len: Size> SimValue<ArrayType<T, Len>> {
    #[track_caller]
    pub fn from_array_elements<I: IntoIterator<Item: ToSimValue<T>>>(
        ty: ArrayType<T, Len>,
        elements: I,
    ) -> Self {
        let element_ty = ty.element();
        let elements = Vec::from_iter(
            elements
                .into_iter()
                .map(|element| element.into_sim_value(element_ty)),
        );
        assert_eq!(elements.len(), ty.len());
        SimValue::from_value(ty, elements.try_into().ok().expect("already checked len"))
    }
}

impl<Element: ToSimValue<T>, T: Type> ToSimValue<Array<T>> for [Element] {
    #[track_caller]
    fn to_sim_value(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<CanonicalType>> ToSimValue<CanonicalType> for [Element] {
    #[track_caller]
    fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<Element: ToSimValue<T>, T: Type, const N: usize> ToSimValue<Array<T, N>> for [Element; N]
where
    ConstUsize<N>: KnownSize,
{
    #[track_caller]
    fn to_sim_value(&self, ty: Array<T, N>) -> SimValue<Array<T, N>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value(self, ty: Array<T, N>) -> SimValue<Array<T, N>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<T>, T: Type, const N: usize> ToSimValue<Array<T>> for [Element; N] {
    #[track_caller]
    fn to_sim_value(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value(self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<CanonicalType>, const N: usize> ToSimValue<CanonicalType>
    for [Element; N]
{
    #[track_caller]
    fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
    #[track_caller]
    fn into_sim_value(self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<Element: ToSimValue<T>, T: Type> ToSimValue<Array<T>> for Vec<Element> {
    #[track_caller]
    fn to_sim_value(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value(self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<CanonicalType>> ToSimValue<CanonicalType> for Vec<Element> {
    #[track_caller]
    fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
    #[track_caller]
    fn into_sim_value(self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<T: Type> ToSimValue<T> for Expr<T> {
    #[track_caller]
    fn to_sim_value(&self, ty: T) -> SimValue<T> {
        assert_eq!(Expr::ty(*self), ty);
        SimValue::from_bitslice(
            ty,
            &crate::expr::ToLiteralBits::to_literal_bits(self)
                .expect("must be a literal expression"),
        )
    }
}

macro_rules! impl_to_sim_value_for_bool_like {
    ($ty:ident) => {
        impl ToSimValue<$ty> for bool {
            fn to_sim_value(&self, ty: $ty) -> SimValue<$ty> {
                SimValue::from_value(ty, *self)
            }
        }
    };
}

impl_to_sim_value_for_bool_like!(Bool);
impl_to_sim_value_for_bool_like!(AsyncReset);
impl_to_sim_value_for_bool_like!(SyncReset);
impl_to_sim_value_for_bool_like!(Reset);
impl_to_sim_value_for_bool_like!(Clock);

impl ToSimValue<CanonicalType> for bool {
    #[track_caller]
    fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        match ty {
            CanonicalType::UInt(_)
            | CanonicalType::SInt(_)
            | CanonicalType::Array(_)
            | CanonicalType::Enum(_)
            | CanonicalType::Bundle(_)
            | CanonicalType::PhantomConst(_) => {
                panic!("can't create SimValue from bool: expected value of type: {ty:?}");
            }
            CanonicalType::Bool(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_)
            | CanonicalType::Clock(_) => {
                SimValue::from_bits(ty, UIntValue::new(Arc::new(BitVec::repeat(*self, 1))))
            }
        }
    }
}

macro_rules! impl_to_sim_value_for_primitive_int {
    ($prim:ident) => {
        impl ToSimValue<<$prim as ToExpr>::Type> for $prim {
            #[track_caller]
            fn to_sim_value(
                &self,
                ty: <$prim as ToExpr>::Type,
            ) -> SimValue<<$prim as ToExpr>::Type> {
                SimValue::from_value(ty, (*self).into())
            }
        }

        impl ToSimValue<<<$prim as ToExpr>::Type as IntType>::Dyn> for $prim {
            #[track_caller]
            fn to_sim_value(
                &self,
                ty: <<$prim as ToExpr>::Type as IntType>::Dyn,
            ) -> SimValue<<<$prim as ToExpr>::Type as IntType>::Dyn> {
                SimValue::from_value(
                    ty,
                    <<$prim as ToExpr>::Type as Type>::SimValue::from(*self).as_dyn_int(),
                )
            }
        }

        impl ToSimValue<CanonicalType> for $prim {
            #[track_caller]
            fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
                let ty: <<$prim as ToExpr>::Type as IntType>::Dyn = Type::from_canonical(ty);
                SimValue::into_canonical(self.to_sim_value(ty))
            }
        }
    };
}

impl_to_sim_value_for_primitive_int!(u8);
impl_to_sim_value_for_primitive_int!(u16);
impl_to_sim_value_for_primitive_int!(u32);
impl_to_sim_value_for_primitive_int!(u64);
impl_to_sim_value_for_primitive_int!(u128);
impl_to_sim_value_for_primitive_int!(usize);
impl_to_sim_value_for_primitive_int!(i8);
impl_to_sim_value_for_primitive_int!(i16);
impl_to_sim_value_for_primitive_int!(i32);
impl_to_sim_value_for_primitive_int!(i64);
impl_to_sim_value_for_primitive_int!(i128);
impl_to_sim_value_for_primitive_int!(isize);

macro_rules! impl_to_sim_value_for_int_value {
    ($IntValue:ident, $Int:ident, $IntType:ident) => {
        impl<Width: KnownSize> ToSimValue<$IntType<Width>> for $IntValue<Width> {
            fn to_sim_value(&self, ty: $IntType<Width>) -> SimValue<$IntType<Width>> {
                self.bits().to_sim_value(ty)
            }

            fn into_sim_value(self, ty: $IntType<Width>) -> SimValue<$IntType<Width>> {
                self.into_bits().into_sim_value(ty)
            }
        }

        impl<Width: Size> ToSimValue<$Int> for $IntValue<Width> {
            fn to_sim_value(&self, ty: $Int) -> SimValue<$Int> {
                self.bits().to_sim_value(ty)
            }

            fn into_sim_value(self, ty: $Int) -> SimValue<$Int> {
                self.into_bits().into_sim_value(ty)
            }
        }

        impl<Width: Size> ToSimValue<CanonicalType> for $IntValue<Width> {
            #[track_caller]
            fn to_sim_value(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
                SimValue::into_canonical(self.to_sim_value($Int::from_canonical(ty)))
            }

            #[track_caller]
            fn into_sim_value(self, ty: CanonicalType) -> SimValue<CanonicalType> {
                SimValue::into_canonical(self.into_sim_value($Int::from_canonical(ty)))
            }
        }
    };
}

impl_to_sim_value_for_int_value!(UIntValue, UInt, UIntType);
impl_to_sim_value_for_int_value!(SIntValue, SInt, SIntType);
